"""Custom build backend to download Sensapex SDK binaries during installation.
This extends setuptools.build_meta to add download hooks."""

from pathlib import Path
from setuptools import build_meta as _orig

# Re-export all the standard hooks
__all__ = [
    "build_wheel",
    "build_sdist",
    "build_editable",
    "get_requires_for_build_wheel",
    "get_requires_for_build_sdist",
    "get_requires_for_build_editable",
    "prepare_metadata_for_build_wheel",
    "prepare_metadata_for_build_editable",
]


def _download_binaries(wheel_directory, config_settings):
    """Download SDK binaries after build."""
    from sensapex._build import install_bin

    # Install binaries to the sensapex package directory
    source_dir = Path(__file__).parent / "sensapex"
    print(f"Downloading Sensapex SDK binaries to {source_dir}")
    install_bin(source_dir)


# Expose standard build backend functions
get_requires_for_build_wheel = _orig.get_requires_for_build_wheel
get_requires_for_build_sdist = _orig.get_requires_for_build_sdist
prepare_metadata_for_build_wheel = _orig.prepare_metadata_for_build_wheel


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Build a wheel and include downloaded binaries."""
    _download_binaries(wheel_directory, config_settings)
    return _orig.build_wheel(wheel_directory, config_settings, metadata_directory)


def build_sdist(sdist_directory, config_settings=None):
    """Build a source distribution."""
    return _orig.build_sdist(sdist_directory, config_settings)


# Handle editable installs
if hasattr(_orig, "get_requires_for_build_editable"):
    get_requires_for_build_editable = _orig.get_requires_for_build_editable
    prepare_metadata_for_build_editable = _orig.prepare_metadata_for_build_editable

    def build_editable(wheel_directory, config_settings=None, metadata_directory=None):
        """Build an editable install and include downloaded binaries."""
        _download_binaries(wheel_directory, config_settings)
        return _orig.build_editable(wheel_directory, config_settings, metadata_directory)



"""Utilities for downloading required Sensapex SDK binaries."""

from __future__ import annotations

import os
from io import BytesIO
import platform
from pathlib import Path
from typing import Iterable, List
from urllib.parse import urlparse
from zipfile import ZipFile
import urllib.request


if platform.system() == "Windows":
    UMSDK_URL = "https://github.com/sensapex/umsdk/releases/download/v1.504/umsdk_Windows_X64_gcc_v1.504.3_Release.zip"
    UMSDK_MEMBERS = ["bin/um.dll"]
elif platform.system() == "Darwin" and platform.machine() == "arm64":
    UMSDK_URL = "https://github.com/sensapex/umsdk/releases/download/v1.504/umsdk_macOS_ARM64_gcc_v1.504.3_Release.zip"
    UMSDK_MEMBERS = ["bin/shared/libum.dylib"]
elif platform.system() == "Linux" and platform.machine() == "x86_64":
    UMSDK_URL = "https://github.com/sensapex/umsdk/releases/download/v1.504/umsdk_Linux_X64_gcc_v1.504.3_Release.zip"
    UMSDK_MEMBERS = ["bin/shared/libum.so"]
else:
    UMSDK_URL = None  # Unsupported platform

UMSDK_ENV = "SENSAPEX_UMSDK_ARCHIVE"

UMPCLI_URL = "http://dist.sensapex.com/misc/umpcli/umpcli-0_957-beta.zip"
UMPCLI_MEMBERS = ["umpcli.exe"]
UMPCLI_ENV = "SENSAPEX_UMPCLI_ARCHIVE"

CACHE_DIR = Path(
    os.environ.get(
        "SENSAPEX_DRIVER_CACHE",
        Path.home() / ".cache" / "sensapex",
    )
)
REPO_ROOT = Path(__file__).resolve().parent.parent


def install_bin(path: Path) -> None:
    """Install platform-specific libum library (dll/dylib/so) to *path*."""
    # Check if platform is supported
    if UMSDK_URL is None:
        print(f"Warning: Sensapex SDK is not available for {platform.system()} {platform.machine()}")
        return

    path.mkdir(parents=True, exist_ok=True)

    # Download and extract the platform-specific library
    dll_data = download_from_zip(UMSDK_URL, UMSDK_MEMBERS, env_var=UMSDK_ENV)[0]

    # Determine the output filename based on platform
    if platform.system() == "Windows":
        lib_filename = "um.dll"
    elif platform.system() == "Darwin":
        lib_filename = "libum.dylib"
    else:  # Linux
        lib_filename = "libum.so"

    (path / lib_filename).write_bytes(dll_data)

    # Only download umpcli on Windows
    if platform.system() == "Windows":
        umpcli_data = download_from_zip(UMPCLI_URL, UMPCLI_MEMBERS, env_var=UMPCLI_ENV)[0]
        (path / "umpcli.exe").write_bytes(umpcli_data)


def download_from_zip(url: str, files: List[str], env_var: str | None = None) -> List[bytes]:
    content_file = BytesIO(_get_archive_bytes(url, env_var))
    data = []
    with ZipFile(content_file, "r") as zip_file:
        for filename in files:
            with zip_file.open(filename) as req_file:
                data.append(req_file.read())
    return data


def _get_archive_bytes(url: str, env_var: str | None) -> bytes:
    filename = Path(urlparse(url).path).name
    for candidate in _archive_sources(filename, url, env_var):
        if isinstance(candidate, Path):
            if candidate.is_file():
                return candidate.read_bytes()
            raise FileNotFoundError(f"Configured archive {candidate} does not exist.")

        data = _download_url(candidate)
        CACHE_DIR.mkdir(parents=True, exist_ok=True)
        (CACHE_DIR / filename).write_bytes(data)
        return data

    raise RuntimeError(f"Unable to fetch archive for {url}")


def _archive_sources(filename: str, url: str, env_var: str | None) -> Iterable[Path | str]:
    if env_var:
        explicit = os.environ.get(env_var)
        if explicit:
            yield Path(explicit).expanduser()

    repo_candidate = REPO_ROOT / filename
    if repo_candidate.exists():
        yield repo_candidate

    cache_candidate = CACHE_DIR / filename
    if cache_candidate.exists():
        yield cache_candidate

    yield url


def _download_url(url: str) -> bytes:
    try:
        with urllib.request.urlopen(url, timeout=60) as req:
            return req.read()
    except Exception as exc:  # pragma: no cover - best effort error reporting
        raise RuntimeError(f"Unable to download {url}: {exc}") from exc



import platform


def bytes_str(s):
    return bytes(s, "utf-8")


packet_count_param = "-n" if platform.system().lower() == "windows" else "-c"


# def ping(host, timeout=1):
#     """Ping *host* once and return the latency, or None if no response is received.
#     """
#     timeout = int(timeout)
#     try:
#         now = time.time()
#         addr = get_host_ip(host)
#         out = subprocess.check_output(['ping', '-c1', '-n', addr], timeout=timeout).decode()
#         if time.time() - now > 1:
#             print("!!!!", host)
#             print(out)
#     except subprocess.TimeoutExpired:
#         return None
#     except Exception as exc:
#         raise Exception("Ping %s failed: %s" % (host, exc.output.decode()))
#
#     n_recv = None
#     for line in out.split('\n'):
#         m = re.match(r'(\d) packets transmitted, (\d) received.*', line)
#         if m is not None:
#             n_sent = int(m.groups()[0])
#             n_recv = int(m.groups()[1])
#             if n_recv < n_sent:
#                 print(host, 'timeout')
#                 return None
#         m = re.match(r'rtt min/avg/max/mdev = (?P<min>[0-9\.]+)/(?P<avg>[0-9\.]+)/(?P<max>[0-9\.]+)/.*', line)
#         if m is not None:
#             if n_recv is None:
#                 break
#             lat = float(m.group('max'))
#             return lat * 1e-3
#
#     raise Exception("Could not parse ping output: %s" % out)
#
#
# def ping_parallel(addrs, timeout=3, max_workers=16):
#     """Ping multiple addresses in parallel.
#     """
#     exec = concurrent.futures.ThreadPoolExecutor(max_workers=min(max_workers, len(addrs)))
#     futures = {}
#     for addr in addrs:
#         futures[addr] = exec.submit(ping, addr, timeout)
#     return {addr: fut.result() for (addr, fut) in futures.items()}
#
#
# class TrioThread(threading.Thread):
#     def __init__(self, addresses: Iterable[str], on_complete):
#         self._addresses = addresses
#         self._on_complete = on_complete
#         self._stop_called = False
#         super(TrioThread, self).__init__(daemon=True)
#
#     def run(self):
#         self._on_complete(trio.run(self._async_scan_subnet_trio))
#
#     def stop(self):
#         self._stop_called = True
#
#     async def _async_scan_subnet_trio(self) -> Set[str]:
#         responsive = set()
#         pings_in_progress = 0
#
#         async def do_ping(host: str, cancel_scope: trio.CancelScope):
#             nonlocal pings_in_progress
#             command = ["ping", packet_count_param, "1", host]
#             while pings_in_progress >= 10:
#                 if self._stop_called:
#                     cancel_scope.cancel()
#                 await trio.sleep(0.1)
#             with trio.move_on_after(1):
#                 pings_in_progress += 1
#                 process = await trio.run_process(
#                     command,
#                     check=False,
#                     stdout=subprocess.DEVNULL,
#                     stderr=subprocess.DEVNULL,
#                 )
#                 if process.returncode == 0:
#                     responsive.add(host)
#             pings_in_progress -= 1
#
#         async with trio.open_nursery() as nursery:
#             self._nursery = nursery
#             [nursery.start_soon(do_ping, h, nursery.cancel_scope) for h in self._addresses]
#
#         return responsive
#
#
# class ScanThread(threading.Thread):
#     def __init__(self, addresses: Iterable[str], on_complete):
#         self._addresses = addresses
#         self._on_complete = on_complete
#         self._stop_called = False
#         super(ScanThread, self).__init__(daemon=True)
#
#     def stop(self):
#         self._stop_called = True
#
#     def run(self):
#         responsive = set()
#         pings_in_progress = 0
#         lock = threading.Lock()
#
#         def do_ping(host: str):
#             nonlocal pings_in_progress
#             command = ["ping", packet_count_param, "1", host]
#             while pings_in_progress >= 10:
#                 if self._stop_called:
#                     return
#                 time.sleep(0.1)
#             with lock:
#                 pings_in_progress += 1
#             try:
#                 process = subprocess.run(command, timeout=1, capture_output=True)
#                 if process.returncode == 0:
#                     responsive.add(host)
#             except subprocess.TimeoutExpired:
#                 pass
#             with lock:
#                 pings_in_progress -= 1
#
#         threads = [threading.Thread(target=do_ping, args=(h,)) for h in self._addresses]
#         [th.start() for th in threads]
#         for th in threads:
#             if not self._stop_called:
#                 th.join()
#         if not self._stop_called:
#             self._on_complete(responsive)



import argparse
import sys
import time

import numpy as np
import pyqtgraph as pg
from sensapex import UMP
from sensapex.sensapex import LIBUM_DEF_BCAST_ADDRESS
from sensapex.utils import bytes_str


parser = argparse.ArgumentParser(
    description="Test for sensapex pressure controllers."
)
parser.add_argument("device", type=int, help="Device ID to test")
parser.add_argument(
    "--library-path", "-l", type=str, dest="library_path", default=None, help="Folder containing the umsdk library"
)
parser.add_argument("--address", "-a", type=bytes_str, default=LIBUM_DEF_BCAST_ADDRESS, help="Device network address")
parser.add_argument("--debug", "-d", default=False, action="store_true", help="Turn on debug logging")
parser.add_argument("--group", "-g", type=int, default=0, help="Device group number")
args = parser.parse_args()

UMP.set_library_path(args.library_path)
ump = UMP.get_ump(address=args.address, group=args.group)
if args.debug:
    try:
        ump.set_debug_mode(True)
    except Exception as e:
        print(f"Could not enable Sensapex debug mode: {e}")
time.sleep(2)
devids = ump.list_devices()
devs = {i: ump.get_device(i) for i in devids}

print("SDK version:", ump.sdk_version())
print("Found device IDs:", devids)

dev = devs[args.device]


app = pg.mkQApp()
win = pg.GraphicsLayoutWidget()
plots = [win.addPlot(row=i, col=0, labels={"left": (f"channel {i+1} (kPa)"), "bottom": ("time", "s")}) for i in range(8)]
for i in range(0, 8):
    plots[i].setXLink(plots[0])
    plots[i].addLine(x=0)
err_plots = [win.addPlot(row=i, col=1) for i in range(8)]
err_curves = [plt.plot(symbol='o') for plt in err_plots]
for i in range(8):
    err_plots[i].addLine([0, 0], angle=45)

win.resize(600, 1000)
win.show()


for chan in range(1, 9):
    start_valve = dev.get_valve(chan)

    dev.set_valve(chan, 0)
    assert dev.get_valve(chan) == 0, f"channel {chan} failed valve test"

    dev.set_valve(chan, 1)
    assert dev.get_valve(chan) == 1, f"channel {chan} failed valve test"

    dev.set_valve(chan, start_valve)

    accuracy = np.inf

    pressure_values = range(-50, 51, 10)
    results = np.zeros((len(pressure_values), 2))
    results[:] = np.nan
    pre_samples = 30
    post_samples = 40

    for i, target_pressure in enumerate(pressure_values):
        pressure_data = []
        time_vals = []
        dev.set_pressure(chan, 0.0)
        for j in range(pre_samples):
            pressure_data.append(dev.measure_pressure(chan))
            time_vals.append(time.perf_counter())
            time.sleep(0.001)
        set_time = time.perf_counter()
        dev.set_pressure(chan, target_pressure)
        for j in range(post_samples):
            pressure_data.append(dev.measure_pressure(chan))
            time_vals.append(time.perf_counter())
            time.sleep(0.001)
        pressure_data = np.array(pressure_data)
        results[i] = [target_pressure, pressure_data[-10:].mean()]
        if target_pressure != 0:
            accuracy = min(accuracy, 1.0 - (((np.array(pressure_data[-10:]) - target_pressure)**2).mean()**0.5) / target_pressure)
        plots[chan-1].plot(np.array(time_vals) - set_time, pressure_data)
        err_curves[chan-1].setData(results[:,0], results[:,1], symbol='o')
        app.processEvents()
    dev.set_pressure(chan, 0.0)

    app.processEvents()

    print(f"Channel {chan} pressure accuracy: {accuracy}")




if sys.flags.interactive == 0:
    app.exec_()




import argparse
import sys
import time

import numpy as np
import pyqtgraph as pg
from sensapex import UMP
from sensapex.sensapex import LIBUM_DEF_BCAST_ADDRESS
from sensapex.utils import bytes_str


parser = argparse.ArgumentParser(
    description="Test for sensapex devices; perform a series of random moves while rapidly polling the device position"
                " and state."
)
parser.add_argument("device", type=int, help="Device ID to test")
parser.add_argument(
    "--library-path", "-l", type=str, dest="library_path", default=None, help="Folder containing the umsdk library"
)
parser.add_argument("--address", "-a", type=bytes_str, default=LIBUM_DEF_BCAST_ADDRESS, help="Device network address")
parser.add_argument("--debug", "-d", action="store_true", help="Turn on debug logging")
parser.add_argument("--group", "-g", type=int, default=0, help="Device group number")
parser.add_argument(
    "--x", action="store_true", default=False, dest="x", help="True = Random X axis values. False = keep start position"
)
parser.add_argument(
    "--y", action="store_true", default=False, dest="y", help="True = Random Y axis values. False = keep start position"
)
parser.add_argument(
    "--z", action="store_true", default=False, dest="z", help="True = Random Z axis values. False = keep start position"
)

parser.add_argument("--speed", type=float, default=1000, help="Movement speed in um/sec")
parser.add_argument(
    "--distance", type=float, default=10, help="Max distance to travel in um (relative to current position)"
)
parser.add_argument("--iter", type=int, default=10, help="Number of positions to test")
parser.add_argument("--acceleration", type=float, default=0, help="Max speed acceleration")
parser.add_argument(
    "--retry-threshold",
    type=float,
    default=None,
    dest="retry_threshold",
    help="Distance error threshold at which to retry a move",
)
parser.add_argument(
    "--linear", action="store_true", default=False, dest="linear", help="Move all 3 axes simultaneously"
)
parser.add_argument(
    "--high-res",
    action="store_true",
    default=False,
    dest="high_res",
    help="Use high-resolution time sampling rather than poller's schedule",
)
parser.add_argument(
    "--start-pos",
    type=str,
    default=None,
    dest="start_pos",
    help="x,y,z starting position (by default, the current position is used)",
)
parser.add_argument(
    "--test-pos",
    type=str,
    default=None,
    dest="test_pos",
    help="x,y,z position to test (by default, random steps from the starting position are used)",
)
args = parser.parse_args()

UMP.set_library_path(args.library_path)
ump = UMP.get_ump(address=args.address, group=args.group)

if args.debug:
    try:
        ump.set_debug_mode(True)
    except Exception as e:
        print(f"Could not enable Sensapex debug mode: {e}")
time.sleep(2)
devids = ump.list_devices()
devs = {i: ump.get_device(i) for i in devids}

print("SDK version:", ump.sdk_version())
print("Found device IDs:", devids)

dev = devs[args.device]

app = pg.mkQApp()
win = pg.GraphicsLayoutWidget()
win.show()
plots = [
    win.addPlot(labels={"left": ("x position", "m"), "bottom": ("time", "s")}),
    win.addPlot(labels={"left": ("y position", "m"), "bottom": ("time", "s")}),
    win.addPlot(labels={"left": ("z position", "m"), "bottom": ("time", "s")}),
]
plots[1].setYLink(plots[0])
plots[2].setYLink(plots[0])
plots[1].setXLink(plots[0])
plots[2].setXLink(plots[0])

win.nextRow()
errplots = [
    win.addPlot(labels={"left": ("x error", "m"), "bottom": ("time", "s")}),
    win.addPlot(labels={"left": ("y error", "m"), "bottom": ("time", "s")}),
    win.addPlot(labels={"left": ("z error", "m"), "bottom": ("time", "s")}),
]
errplots[1].setYLink(errplots[0])
errplots[2].setYLink(errplots[0])
errplots[0].setXLink(plots[0])
errplots[1].setXLink(plots[0])
errplots[2].setXLink(plots[0])

if args.linear:
    win.nextRow()
    linerrplots = [
        win.addPlot(labels={"left": ("x linear error", "m"), "bottom": ("time", "s")}),
        win.addPlot(labels={"left": ("y linear error", "m"), "bottom": ("time", "s")}),
        win.addPlot(labels={"left": ("z linear error", "m"), "bottom": ("time", "s")}),
    ]
    linerrplots[1].setYLink(linerrplots[0])
    linerrplots[2].setYLink(linerrplots[0])
    linerrplots[0].setXLink(plots[0])
    linerrplots[1].setXLink(plots[0])
    linerrplots[2].setXLink(plots[0])

start = time.perf_counter()
pos = [[], [], []]
tgt = [[], [], []]
err = [[], [], []]
closest = [[], [], []]
linear_err = [[], [], []]
bus = []
mov = []
times = []

lastupdate = time.perf_counter()


def update(moving=True):
    global lastupdate, n_axes
    timeout = -1 if args.high_res else 0
    position = dev.get_pos(timeout=timeout)
    s = dev.is_busy()
    m = not move_req.finished
    bus.append(int(s))
    mov.append(int(m))
    now = time.perf_counter() - start
    times.append(now)

    # calculate closest point to the line from starting position to target
    target_to_pos = position - target
    target_to_last = last_position_before_move - target
    target_to_last /= np.linalg.norm(target_to_last)
    closest_pos = target + np.dot(target_to_pos, target_to_last) * target_to_last
    dist = position - closest_pos

    for i in range(min(3, n_axes)):
        pos[i].append((position[i] - start_pos[i]) * 1e-6)
        tgt[i].append((target[i] - start_pos[i]) * 1e-6)
        if moving:
            err[i].append(np.nan)
            # only update linear error when moving
            closest[i].append((closest_pos[i] - start_pos[i]) * 1e-6)
            linear_err[i].append(dist[i] * 1e-6)
        else:
            # only update position error when stopped
            err[i].append(pos[i][-1] - tgt[i][-1])
            closest[i].append(np.nan)
            linear_err[i].append(np.nan)


def update_plots():
    for i in range(3):
        plots[i].clear()
        plots[i].addItem(
            pg.PlotCurveItem(times, bus[:-1], stepMode=True, pen=None, brush=(0, 255, 0, 40), fillLevel=0),
            ignoreBounds=True,
        )
        plots[i].addItem(
            pg.PlotCurveItem(times, mov[:-1], stepMode=True, pen=None, brush=(255, 0, 0, 40), fillLevel=0),
            ignoreBounds=True,
        )
        plots[i].plot(times, tgt[i], pen="r")
        plots[i].plot(times, pos[i], symbol="o", symbolSize=5)
        plots[i].plot(times, closest[i], pen=(100, 100, 100))
        errplots[i].plot(times, err[i], clear=True, connect="finite")
        if args.linear:
            linerrplots[i].plot(times, linear_err[i], clear=True, connect="finite")


if args.start_pos is None:
    start_pos = dev.get_pos()
else:
    start_pos = np.array(list(map(float, args.start_pos.split(","))))

print("Starting position:", start_pos)
diffs = []
errs = []
positions = []
n_axes = dev.n_axes()
assert (
    len(start_pos) == n_axes
), f"Starting position {start_pos} has length {len(start_pos)}, but device has {n_axes} axes."
if args.test_pos is None:
    moves = np.random.random(size=(args.iter, n_axes)) * args.distance
    move_axes = np.array([args.x, args.y, args.z, False])[:n_axes]
    assert np.any(move_axes), "No axes selected to move (use --x, --y, --z, or --test-pos)"
    moves[:, ~move_axes] = 0
    targets = np.array(start_pos)[np.newaxis, :] + moves
    print(f"Distance to move each axis:\n{moves}")
    print(f"Target positions:\n{targets}")
else:
    # just move back and forth between start and test position
    test_pos = np.array(list(map(float, args.test_pos.split(","))))
    targets = np.zeros((args.iter, 3))
    targets[::2] = start_pos[None, :]
    targets[1::2] = test_pos[None, :]
speeds = [args.speed] * args.iter


dev.stop()

if args.retry_threshold is not None:
    ump.set_retry_threshold(args.retry_threshold)

for i in range(args.iter):
    target = targets[i]
    last_position_before_move = dev.get_pos()
    move_req = dev.goto_pos(
        target, speed=speeds[i], linear=args.linear, simultaneous=args.linear, max_acceleration=args.acceleration
    )
    while not move_req.finished:
        update(moving=True)
        time.sleep(0.002)
    waitstart = time.perf_counter()
    while time.perf_counter() - waitstart < 1.0:
        update(moving=False)
        time.sleep(0.002)
        # time.sleep(0.05)
    p2 = dev.get_pos(timeout=200)
    positions.append(p2)
    diff = (p2 - target) * 1e-6
    diffs.append(diff)
    errs.append(np.linalg.norm(diff))

    print(
        f"{i} attempts: {move_req.attempts}  error: {errs[-1]*1e6:0.2f} um "
        f"  [{' '.join(['%0.2f'%(x*1e6) for x in diff])}]"
    )

update_plots()

dev.goto_pos(start_pos, args.speed)
print(f"mean: {np.mean(errs) * 1e6:0.2f} um   max: {np.max(errs) * 1e6:0.2f} um")

if sys.flags.interactive == 0:
    app.exec_()
